# -*- coding: utf-8 -*-
'''
Taint Mode for Python via a Library


Copyright 2009 Juan José Conti
Copyright 2010 Juan José Conti - Alejandro Russo

This file is part of taintmode.py

taintmode is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

taintmode.py is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with taintmode.py.  If not, see <http://www.gnu.org/licenses/>.

'''
import sys
from itertools import chain
import types
import taintConfig
from Policies import TaintException

sys.path.insert(0, 'google_appengine')  # This part is need to access to google.appengine.ext.db
import dev_appserver                    # what is needed to know about db.*Property datatypes
dev_appserver.fix_sys_path()            #
from google.appengine.ext import db     #

__version__ = 'trunk-svn-2'

#__all__ = ['tainted', 'deeptainted', 'taint', 'untrusted', 'untrusted_args', 'ssink',
#           'validator', 'cleaner', 'STR', 'INT', 'FLOAT', 'UNICODE', 'chr',
#           'ord', 'len', 'ends_execution', 'taintModel','untaint_aware','DtStringIO','DtString']


ENDS = False
RAISES = False

import pickle

def taintModel(originalModelType):
    class Callable: #CHECKME little wrapper to create class-methods. Possibly, there is a better way to do this.
        def __init__(self, anycallable):
            self.__call__ = anycallable

    class tModel(originalModelType):
        taints=db.TextProperty()

        def __init__(self,*args, **kargs):
           #TODO Possibly, this method need to be refactored
           taints={}
           for key,value in kargs.iteritems():
               if tainted(value):
                   value_taints=collect_tags(value)
                   taints[key]=value_taints
                   kargs[key]=untaint_aware(value)

           # This conditional is for the get cases. get needs to contruct a tModel. Do not overwrite model.taints with empty in that case
           if kargs.has_key('taints'):
               dict_taints=pickle.loads(str(kargs['taints']))
               _taints=kargs['taints']
               del kargs['taints']
               for property,ptaints in dict_taints.iteritems():
                   kargs[property]=taint_aware(kargs[property],ptaints)
           else:
               _taints=pickle.dumps(taints)

           super(tModel, self).__init__(*args, **kargs)
           self.taints=_taints

        def updateTaintsFor(self, property, taints):
            '''Update (overwrite) the taints of the the property'''
            dict_taints=pickle.loads(str(self.taints))
            dict_taints[property]=taints
            self.taints=pickle.dumps(dict_taints)

        def __setattr__(self, name, value):
            if name in self.properties().keys() and name != 'taints':
                # similiar code in the __init__ TODO refactor
                taints=collect_tags(value)
                originalModelType.__setattr__(self, name, untaint_aware(value))
                self.updateTaintsFor(name, taints)
            else:
                # Do nothing, because is something internal
                originalModelType.__setattr__(self, name, value)

        def get_wrapper(*args, **kargs):
            model=originalModelType.get(*args, **kargs)
            if hasattr(model,'taints'):
                taints=pickle.loads(str(model.taints)) # the str cast is needed, because db converts string in unicode
                for property,taint in taints.iteritems():
                    #convert that value in a taint_aware value
                    originalModelType.__setattr__(model, property, taint_aware(getattr(model,property),taint))
            return model

        get=Callable(get_wrapper)

    return tModel

def ends_execution(b=True):
    global ENDS
    ENDS = b


# ------------------------- Auxiliaries functions -----------------------------

# ------------------------- Decorators ----------------------------------------

def untrusted_args(nargs=[], nkwargs=[]):
    '''
    Mark a function or method that would recive untrusted values.

    nargs is a list of positions. Positional arguments in that position will be
    tainted for all the types of taint.

    nkwargs is a list of strings. Keyword arguments for those keys will be
    tainted for all the types of taint.

    Some frameworks works as follow: they ask the programmer to write certain
    function or method in such a way that then the framework itself use it to
    pass to the program values received from the outside. Twisted, the framework
    for building network applications, gives us an accurate example when you
    extend the LineOnlyReceiver class:

    .. code-block:: python

      class MyProtocol(LineOnlyReceiver):

        def lineReceived(self, line):
            self.doSomething(line)  # line is a tainted value

    It's complicated or even impossible use the untrusted decorator in this kind
    of situations. Instead of it, you should use the untrusted_args decorator,
    which receives as an optional argument a list of non-trusted arguments positions
    and a list of keywords. The line parameter in the example can be marked as
    untrusted in this way:

    .. code-block:: python

        class MyProtocol(LineOnlyReceiver):

        @untrusted_args([1])
        def lineReceived(self, line):
            self.doSomething(line)

    Example
    =======
    >>> import taintmode
    >>> from taintmode import *

    >>> @ssink(OSI)
    ... def exec_comand(cm):
    ...     print "executing ", cm
    ...
    >>> @untrusted_args([1])
    ... def rutine(auxvalue, command):
    ...     exec_comand(command)
    ...     print auxvalue
    ...
    >>> rutine(42,"rm -r /")
    ===============================================================================
    Violation in line 3 from file <doctest __main__.untrusted_args[3]>
    Tainted value: rm -r /
    -------------------------------------------------------------------------------
            print auxvalue
    <BLANKLINE>
    ===============================================================================
    executing  rm -r /
    42
    '''
    def _untrusted_args(f):
        def inner(*args, **kwargs):
            args = list(args)   # args is a tuple - add a test
            for n in nargs:
                args[n] = mapt(args[n], taint)
            for n in nkwargs:
                kwargs[n] = mapt(kwargs[n], taint)
            r = f(*args, **kwargs)
            return r
        return inner
    return _untrusted_args

def untrusted(f):
    '''
    Mark a function or method as untrusted.

    The returned value will be tainted for all the types of taint.

    Examples
    ========

    If you have access to the function or method definition, for example if it's part
    of your code-base the decorator can be applied using Python's syntactic sugar:

    >>> @untrusted
    ... def from_outside():
    ...     return 'a string' #this value is untrusted
    ...
    >>> value = from_outside()
    >>> value
    'a string'
    >>> type(value)
    <class '__main__.tklass'>
    >>> tainted(value)
    True

    While using third-party modules, we still can apply the decorator. The next
    example is from a program writed using the web.py framework:

    >>> import web
    >>> web.input = untrusted(web.input)
    '''
    def inner(*args, **kwargs):
        r = f(*args, **kwargs)
        return taint(r)
    return inner

def validator(v, cond=True, nargs=[], nkwargs=[]):
    '''
    Mark a function or method as capable to validate its input.

    nargs is a list of positions. Positional arguments in that positions are
    the ones validated.
    nkwargs is a list of strings. Keyword arguments for those keys are the ones
    validated.
    If the function returns cond, v will be removed from the the validated
    inpunt.

    Example:

    for a function called invalid_mail, cond is liked to be False. If
    invalid_mail returns False, then the mail IS valid and have no craft data
    on it.

    for a function called valid_mail, cond is liked to be True.
    '''
    def _validator(f):
        def inner(*args, **kwargs):
            r = f(*args, **kwargs)
            if r == cond:
                tovalid = set(args[n] for n in nargs)
                tovalid.update(kwargs[n] for n in nkwargs)
                for a in tovalid:
                    remove_tags(a, v)
            return r
        return inner
    return _validator


def cleaner(v):
    '''
    Mark a function or methos as capable to clean its input.

    :param v: taint removed from the returned value.

    Example
    =======

    >>> @cleaner(SQLI) #this make the following function capable to clean objets from SQLI taints
    ... def clean_string(stri):
    ...     return stri.split(' ')[0]
    ...
    >>> @untrusted #simulates an utrusted value
    ... def get_id_employee():
    ...     return "21 or 1=1"
    ...

    >>> @ssink(SQLI) #simulate a rutine that will execute an sql query
    ... def erase_employee(id):
    ...     print "this value this value is placed in the WHERE clause of a sql delete statment:" , id
    >>> i=get_id_employee()
    >>> erase_employee(i)
    ===============================================================================
    Violation in line 1 from file <doctest __main__.cleaner[4]>
    Tainted value: 21 or 1=1
    -------------------------------------------------------------------------------
    --> erase_employee(i)
    <BLANKLINE>
    ===============================================================================
    this value this value is placed in the WHERE clause of a sql delete statment: 21 or 1=1

    >>> i = clean_string(i) # the untrusted value is clean now
    >>> i
    '21'
    >>> erase_employee(i)
    this value this value is placed in the WHERE clause of a sql delete statment: 21
    '''
    def _cleaner(f):
        def inner(*args, **kwargs):
            r = f(*args, **kwargs)
            remove_tags(r, v)
            return r
        return inner
    return _cleaner

def reached(t, v=None):
    '''
    Execute if a tainted value reaches a sensitive sink
    for the vulnerability v.

    If the module-level variable ENDS is set to True, then the sink is not
    executed and the reached function is executed instead. If ENDS is set to False,
    the reached function is executed but the program continues its flow.

    The provided de facto implementation alerts that the violation happened and
    information to find the error.
    '''
    frame = sys._getframe(3)
    filename = inspect.getfile(frame)
    lno = frame.f_lineno
    ret= "=" * 79
    ret+= "Violation in line %d from file %s" % (lno, filename)
    # Localize this message
    ret+= "Tainted value: %s" % t
    ret+= '-' * 79
    lineas = inspect.findsource(frame)[0]
    lineas = ['    %s' % l for l in lineas]
    lno = lno - 1
    lineas[lno] = '--> ' + lineas[lno][4:]
    lineas = lineas[lno - 3: lno + 3]
    ret+= "".join(lineas)
    ret+= "=" * 79
    print ret

def ssink(v=None, reached=reached):
    '''
    Mark a function or method as sensitive to tainted data.

    If it is called with a value with the v tag
    (or any tag if v is None),
    it's not executed and reached is executed instead.

    These sinks are sensitive to a kind of vulnerability, and must be specified when
    the decorator is used

    Examples
    ========

    The web.py framework offers SQL Injection sensitive sink examples:

    >>> import web
    >>> db = web.database(dbn="sqlite", db="DB_NAME")
    >>> db.delete=ssink(SQLI)(db.delete)
    >>> db.select = ssink(SQLI)(db.select)
    >>> db.insert = ssink(SQLI)(db.insert)

    Like the rest of decorators, if the sensitive sink is defined in our code, we can
    use syntactic sugar:

    @ssink(OSI):
    def list_dir(input)
        ...

    The decorator can also be used without specifying a vulnerability. In this case,
    the sink is marked as sensitive to every kind of vulnerability, although this is not
    a very common use case:

    @ssink():
    def very_sensitive(input):
       ...

    '''
    def _solve(a, f, args, kwargs):
        if ENDS:
            if RAISES:
                reached(a)
                raise TaintException()
            else:
                return reached(a)
        else:
            reached(a)
            return f(*args, **kwargs)

    def _ssink(f):
        def inner(*args, **kwargs):
            allargs = chain(args, kwargs.itervalues())
            if v is None:   # sensitive to ALL
                for a in allargs:
                    t = collect_tags(a)
                    if not t.isBottom:
                        return _solve(a, f, args, kwargs)
            else:
                for a in allargs:
                    t = collect_tags(a)
                    if t.has(v):
                        return _solve(a, f, args, kwargs)
            return f(*args, **kwargs)
        return inner
    return _ssink

#def deeptainted(o):
#    '''
#    Tells if a container object o, has any tainted on it for the given
#    vulnerability v.
#
#    If v is not provided, checks for all taints.
#    '''
#    s = set()
#    mapt(o, lambda o: (s.add(tainted(o) != set()),o)[1])
#    return (True in s)

def tainted(o, v=None):
    '''
    Tells if a value o, a tclass instance, is tainted for the given
    vulnerability v.

    If v is not provided, checks for all taints. If the value is tainted
    with at least one vulnerability, returns True.

    Example
    =======

    >>> t1 = taint(42, OSI)
    >>> t2 = taint(42)
    >>> tainted(t1)
    True
    >>> tainted(t1, OSI)
    True
    >>> tainted(t1, II)
    False
    >>> tainted(t2, II)
    True
    >>> tainted(t2, SQLI)
    True
    >>> tainted(t2)
    True
    '''
    if isinstance(o,DtString) or isinstance(o,DtStringIO):
        for i in o.taint_details.itervalues():
           if (v is None and not i.isBottom) or (i.has(v)): return True
    if not hasattr(o, 'taints'):
        return False
    if v is not None:
        return o.taints.has(v)
    return not o.taints.isBottom

def taint(o, v=None):
    '''
    Helper function for taint the value o with the vulnerability v.

    If v is not provided, taint with all types of taints.

    Example: tainting with all vulnerabilities
    ==========================================

    >>> t = taint(42)
    >>> t.taints
    set([1, 2, 3, 4])
    >>> tainted(t, XSS)
    True
    >>> tainted(t, OSI)
    True
    >>> tainted(t, SQLI)
    True
    >>> tainted(t, II)
    True

    Example: tainting with Interpreter Injection vulnerability
    ==========================================================

    >>> taint(42, II)
    42
    >>> t = taint(42, II)
    >>> t.taints
    set([4])
    >>> tainted(t, II)
    True
    >>> tainted(t, OSI)
    False

    '''
    if v is not None:
        return taint_aware(o, v)
    else:
        return taint_aware_with_top(o)

# ------------------------- Taint-aware classes -------------------------------

def taint_hook(plain_instance):
    taintedAttributesAndMethods=dict((d,taint(getattr(plain_instance,d))) for d in dir(plain_instance))
    return types.InstanceType(plain_instance.__class__,taintedAttributesAndMethods)

from StringIO import StringIO
class DtStringIO(StringIO):
    def __init__(self,buf=''):
        self.taint_details=dict()
        StringIO.__init__(self,buf)
        self.update_taints(buf,0)

    def write(self,arg):
        self.update_taints(arg,self.pos)
        DtString(StringIO.write(self,arg))

    def getvalue(self):
        s=DtString(StringIO.getvalue(self))
        s.taint_details=self.taint_details
        return s

    def update_taints(self,arg,start):
        # FIXME is the same code than DtString.update_taints!!!
        end=len(arg)+start
        if isinstance(arg,DtString):
            newPositions=dict()
            for pos,taints in arg.taint_details.iteritems():
                if not taints.isBottom: newPositions[(start+pos[0],start+pos[1])]=taints
            self.taint_details.update(newPositions)
        else:
            if tainted(arg):
                if hasattr(arg,'taints') and not arg.taints.isBottom and start!=end:
                    self.taint_details.update({(start,end):arg.taints})
                else: raise Exception('it is tainted, but i dont know what to do with it') #TODO

    def next(self): raise NotImplementedError
    #def close(self): raise NotImplementedError FIXME (maybe remove self.taint_details?)
    def isatty(self): raise NotImplementedError
    #def seek(self, pos, mode = 0): raise NotImplementedError
    def read(self, n = -1): raise NotImplementedError
    def readline(self, length=None): raise NotImplementedError
    def readlines(self, sizehint = 0): raise NotImplementedError
    #def truncate(self, size=None): raise NotImplementedError
    def writelines(self, iterable): raise NotImplementedError

class DtString(str):
#    def __dir__(self):
#        return ['append','read','join']

    def __init__(self,arg=str()):
        if isinstance(arg,DtString):
            self.taint_details=arg.taint_details
        elif hasattr(arg,'taints'):
            self.taint_details={(0,len(arg)):arg.taints}
        else:
            self.taint_details=dict()
        return str.__init__(self, arg)

    def update_taints(self,arg,start):
        end=len(arg)+start
        if isinstance(arg,DtString):
            newPositions=dict()
            for pos,taints in arg.taint_details.iteritems():
                if not taints.isBottom: newPositions[(start+pos[0],start+pos[1])]=taints
            self.taint_details.update(newPositions)
        else:
            if tainted(arg):
                if hasattr(arg,'taints') and not arg.taints.isBottom and start!=end:
                    self.taint_details.update({(start,end):arg.taints})
                else: raise Exception('it is tainted, but i dont know what to do with it') #TODO

    def checkAndAlarm(self,ssink):
	sensitiveTo=taintConfig.Policy.ssinks[ssink]
	subStrings=self.get_tainted_substrings(sensitiveTo)
	if len(subStrings)>0:
	    taintConfig.Policy().alarm(ssink,sensitiveTo,subStrings)

    def get_tainted_substrings(self,taint):
        '''return a list of substrings of self which are tainted (with taint), according with self.taint_details.'''
        ret={}
        for ((taintedSubString_start,taintedSubString_end),_) in filter(lambda x: not x[1],[(i,v.safe_zone(taint)) for i,v in self.taint_details.iteritems()]):
                ret[self[taintedSubString_start:taintedSubString_end]]=self.taint_details[(taintedSubString_start,taintedSubString_end)].state
        return ret

    def join(self,arg):
        offset=0
        newString=DtString(super(DtString,self).join(arg))
        for item in arg[:-1]:
            newString.update_taints(item,offset)
            offset+=len(item)
            newString.update_taints(self,offset)
            offset+=len(self)
        if len(arg)>0:
            newString.update_taints(arg[-1],offset)
        return newString

from taintManagment import remove_tags,collect_tags,update_tags,taint_aware,taint_aware_with_top,untaint_aware,tclass

if __name__ == "__main__":
        import doctest
        doctest.testmod()
